% PARSE_BRK_FILENAME
% Returns the year, month, day of the brk file formats.
% This does not check the validity of the date.
% method - can be 'fsm' or 'filter'. Both should produce same results
%          although the performance will vary.
%
function [yr, mo, dy] = parse_brk_filename(S, method)

if nargin < 2
    method = 'filter';
end

delimiter_count = length(find(S == '_'));

% Delimiter '_' must exist at least twice: yr_mo_dy
if delimiter_count < 2
    disp('Not enough delimiters found.')
    yr = [];
    mo = [];
    dy = [];
    return;
end
    

% Some potentially useful indexes into the string.
prepidx = strfind('prep', S);           % Prep string. May be at beginning or end.
brkidx = strfind('brk_', S);            % Find brk string. Must come before the date.

switch lower(method)
case 'fsm'
            
	% FSM Algorithm:
	% Look for something of the form '##_##_##', where # is a digit using
	% a finite state machine simulator.
	% Save location of matches in case there are more than one.
	% Then we can use brk_ location to determine which one we want.
	% Accepts:
	% 99_99_99
	% 99_99_9933_33_33
	state = 0;
	matches = [];
	for I=1:length(S)
        switch state
        case 0  % No digits matched.
            if isdigit(S(I))
                state = 1;
            end
        case 1  % One digit matched, look for another digit.
            if isdigit(S(I))
                state = 2;
            else state = 0;
            end
        case 2  % Two digits matched, look for delimiter.
            if S(I) == '_'
                state = 3;
            else state = 0;
            end
        case 3  % YR_ matched, look for another digit.
            if isdigit(S(I))
                state = 4;
            else state = 0;
            end
        case 4  % YR_# matched, look for another digit
            if isdigit(S(I))
                state = 5;
            else state = 0;
            end
        case 5  % YR_MO matched, look for delimiter.
            if S(I) == '_'
                state = 6;
            else state = 0;
            end
        case 6  % YR_MO_ matched, look for another digit.
            if isdigit(S(I))
                state = 7;
            else state = 0;
            end
        case 7  % YR_MO_# matched ,look for final digit.
            if isdigit(S(I))    % YR_MO_DY matched.
                matches(end+1) = I-7;   % Store index of beginning of match (7 places back)
            end
            state = 0;          % Always go back to state 0, ignoring what next digit might be.
        end
	end
case 'filter'
    % Convolution algorithm:
    % Use symmetric filter. Convolve/correlate filter with where digits
    % are.
    % Check for the '_' between afterward.
    h = [1 1 -1 1 1 -1 1 1];
    correlation = conv( isdigit(S), h);
    matches = find(correlation == 6);       % If 6 digits matched, we get max correlation of 6.
    matches = matches - length(h) + 1;      % Convolution adds extra length(h)-1 elements.
    for I=1:length(matches) 
        idx = matches(I);
        if S(idx+2) ~= '_' || S(idx+5) ~= '_'
            matches(I) = 0;     % 0 flags to delete
        end
    end
    matches(matches == 0) = []; % Delete flagged entries.
end
    

idx = 0;        % Date index.
if ~isempty(matches)
  	% Find the match that matches the brk_ string.
	if ~isempty(brkidx)
        idx = find(matches == brkidx+4);
        if isempty(idx)
            disp('Date format found, but it is not after brk_')
            idx = matches(1);
        end
	else
        idx = matches(1);
    end
    
    % Add some heuristic checks to distinguish two possible formats.
    % Format types:  
    % 1.  yr_mo_dy
    % 2.  mo_dy_yr 
    
    a = str2double(S(idx:idx+1));
    b = str2double(S(idx+3:idx+4));
    c = str2double(S(idx+6:idx+7));
    
    if a>12 && between(1,b,12) && between(1,c,31)
        yr = a;
        mo = b;
        dy = c;
    elseif between(1,a,12) && between(1,b,31) && c>12
        yr = c;
        mo = a;
        dy = b;
    else
        disp('Unrecognized month,day,year format')
        yr = [];
        mo = [];
        dy = [];
    end
        
else
    disp('Date format (##_##_##) could not be matched.')
    yr = [];
    mo = [];
    dy = [];
end

% ---------------------
% Evaluate inclusive range
function bool = between(a,b,c)
    bool = a <= b & b <= c;

% ---------------------
function bool = isdigit(S)
    bool = '0' <= S & S <= '9';

