function NewFiles = prep_images( fileExpression, varargin )
% PREP_IMAGES automatically performs image operations on a set of image files.
% 
% PREP_IMAGES( fileExpression, varargin )
%  varargin takes in any number of operations to be performed in the
%  order in which it is listed.
%  List of possible operations:
%   'rotate -90'
%   'rotate +90'
%   'mirror y'
%   'mirror x'
%   'invert'
%   'filter' or 'clean'
%
%  fileExpression identifies the image files you with to operate on.
%  fileExpression can contain wildcards such as 
%          '\home\*.tif'
%  Directories and invalid files matching fileExpression will be ignored.
%  PREP_IMAGES creates new files beginning with 'prep-' and will 
%  ignore any files starting with 'prep-' to avoid prepping an alredy
%  prepped file.
%  
%  Rotate will only be performed if image is taller than it is wide.
%  Invert will only be performed if there are more white than black pixels.
%
%  PREP_IMAGES returns a list of files created.
%
%  Example:
%    prep_images('SeisDig/samples/*.tif', 'mirror y', 'clean')
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Steve Chuang
%
% This software is Copyright ? 2002 The Regents of the University of California. 
% All Rights Reserved.
% 
% Permission to use, copy, modify, and distribute this software and its 
% documentation for educational, research and non-profit purposes, without fee, 
% and without a written agreement is hereby granted, provided that the above 
% copyright notice, this paragraph and the following three paragraphs appear in 
% all copies.
% 
% Permission to incorporate this software into commercial products may be
% obtained by contacting Technology Transfer Office, 9500 Gilman Drive, 
% 411 University Center, University of California, La Jolla, CA 92093-0093
% (858) 534-5815, invent@ucsd.edu
% 
% This software program and documentation are copyrighted by The Regents of the
% University of California. The software program and documentation are 
% supplied "as is", without any accompanying services from The Regents. The 
% Regents does not warrant that the operation of the program will be
% uninterrupted 
% or error-free. The end-user understands that the program was developed for 
% research purposes and is advised not to rely exclusively on the program for any 
% reason.
% 
% IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO
% ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
% CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING
% OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
% EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
% THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
% CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
% THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
% THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO
% PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
% MODIFICATIONS.
%
% This software was made with government support under 
% Office of Naval Research Award Number: N000140110403. 
% The government has certain rights in this software.
    
    if nargin < 2
        error('Not enough inputs')
    end

    prefix = reset_prefix;
    % Instead of doing it all here, we'll do it as we go.
%     for n=1:length(varargin)
%         prefix = [prefix nowhite(varargin{n}) '-'];
%     end
    
    NewFiles = [];
    FileStructs = dir(fileExpression);
    [fpath file ext] = fileparts(fileExpression);
    
    for n=1:length(FileStructs)
        % Skip directories
        if FileStructs(n).isdir
            continue
        end
        filename = FileStructs(n).name;
            
        % Check if we've already processed this file

        Duplicate = dir(fullfile(fpath,[prefix filename]));
        
        if ~isempty(Duplicate)
            disp(['Skipping file, ' filename])
            continue    % Ignore already prepped files.
        end

        try
            infile = fullfile(fpath, filename);
            disp(['Opening ' infile])
            I = imread( infile );

            if ndims(I) ~= 2,
                I = uint8(I(:,:,1));                    % Choose only one color channel
            end
            if ~islogical(I),                             % Change image into black & white
                I = im2double(I);
                I = I > (max(I(:))-min(I(:)))/2;             % Pick values greater than middle value
                I = uint8(I);                           % Turn back to uint8
            end
        catch
            disp(['Failed to open ' infile ': ' lasterr])
            %error(lasterr);     % Error out of this function.
            continue
        end
         
        try            
            for n = 1:length(varargin)
                [nrow mcol] = size(I);
                switch lower(deblank(varargin{n}))
                    
                % Rotate ONLY if taller than it is wide
                case {'rotate -90', ... 
                      'rotate counter', ...
                      ...
                      'rotate +90', ...
                      'rotate 90', ...
                      'rotate clockwise', ...
                      'rotate'}
                    switch varargin{n}
                    % Rotate counter clockwise
                    case {'rotate -90', ... 
                          'rotate counter'}
                        rotate_direction = +1;
                    % Rotate clockwise
                    otherwise
                        rotate_direction = -1;
                        
                    end
                    if nrow > mcol
                        rotString = num2str(90*rotate_direction);
                        disp(['Rotating ' rotString ' degrees.'])
                        I = rot90(I,rotate_direction);
                        % Here we update the list of prefix labels.
                        prefix = [prefix 'rot' rotString '_'];
                    else
                        disp(['NOT taller than it is wide. IGNORING ' varargin{n} '!'])
                    end
                case {'mirror y', ...
                      'mirrory', ...
                      'mirror y-axis', ...
                      'mirror yaxis'}
                    disp(['Mirror Y-Axis'])
                    I = I(:,end:-1:1);
                    % Here we update the list of prefix labels.
                     prefix = [prefix 'miry_'];
                case {'mirror x', ...
                      'mirrorx', ...
                      'mirror x-axis', ...
                      'mirror xaxis'}
                    disp(['Mirror X-Axis'])
                    I = I(end:-1:1,:);
                    % Here we update the list of prefix labels.
                    prefix = [prefix 'mirx_'];
                case {'invert', ...
                      'invert image', ...
                      'invert img'}
                    % Invert if more white than black pixels
                    if sum(I(:)) > 0.5*nrow*mcol
                        disp('Inverting...')
                        I = ~I;
                        % Here we update the list of prefix labels.
                        prefix = [prefix 'inv_'];
                    else
                        disp(['More black pixels than white pixels. IGNORING ' varargin{n} '!'])
                    end
                case {'filter', ...
                      'filter image noise',...
                      'filter img noise',...
                      'clean',...
                      'clean img noise',...
                      'clean image noise'}
                    % Filter the image using majority filter
                    disp('Filtering...')
                    I = bwmorph(I,'majority',1);
                    % Here we update the list of prefix labels.
                    prefix = [prefix 'filt_'];
                end
            end% End FOR LOOP
        catch
            warning(['Error processing ' infile '. Skipping.'])
            continue
        end
        
        % Save to disk
        [p f e] = fileparts(filename);
        try
            outpath = fpath; %'/home/stchuang/project/1940/dec/';
            outfile = fullfile(outpath,[ prefix f e]);
            prefix = reset_prefix;
            imwrite(I, outfile, e(2:end))
            disp([outfile ' written as ' e(2:end) ' file.'])
            NewFiles = strvcat(NewFiles,outfile);
        catch
            disp(['Cannot write to directory ''' fpath '''. Writing to current directory.'])
            % If we can't write to target path
            % try writing to current directory
            try
                outfile = ['./' prefix f e];
                if ~isempty(dir(outfile))
                    error(['''' outfile ''' exists. Aborting write.'])
                end
                imwrite(I, outfile, e(2:end))
                disp([outfile ' written as ' e(2:end) ' file.'])
                NewFiles = strvcat(NewFiles,outfile);
            catch
                disp(['Cannot write ' outfile ' to disk: ' lasterr])
            end
        end
    end
    
function prefix = reset_prefix
    PREPEND_1 = 'prep-';
    prefix = PREPEND_1 ;
 
% -------------------------------------
% NOWHITE - remove white space in a string
function string = nowhite(string)
    string = string(:)';    % Make horizontal
    string(isspace(string)) = [];
