% --------------------------------------------------------------------
% SAFE_TRACE - stripped down trace that stops as when it runs
%  into a split, overlap, or gap.
% x0,y0 is the starting point and will NOT be included in data returned.
% [xdata,ydata, endCondition] = safe_trace(x0,y0,Img, RATE, g_width, MERGE, LINE_ENDS_IGNORE, varargin)% varargin{1} = TimeLimit
%     varargin{1} = TimeLimit
%     varargin{2} = DistLimit
%     Defaults:
%         TimeLimit = 0.3;
%         DistLimit = Inf;
%      
%     endCondition = 'gap';       break
%     endCondition = 'endOverlap';   break
%     endCondition = 'distance';  break
%     endCondition = 'timeout';   break
%     endCondition = 'overlap';  break
function [xdata,ydata, endCondition] = safe_trace(x0,y0,Img, RATE, g_width, MERGE, LINE_ENDS_IGNORE, varargin)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Steve Chuang
%
% This software is Copyright ? 2002 The Regents of the University of California. 
% All Rights Reserved.
% 
% Permission to use, copy, modify, and distribute this software and its 
% documentation for educational, research and non-profit purposes, without fee, 
% and without a written agreement is hereby granted, provided that the above 
% copyright notice, this paragraph and the following three paragraphs appear in 
% all copies.
% 
% Permission to incorporate this software into commercial products may be
% obtained by contacting Technology Transfer Office, 9500 Gilman Drive, 
% 411 University Center, University of California, La Jolla, CA 92093-0093
% (858) 534-5815, invent@ucsd.edu
% 
% This software program and documentation are copyrighted by The Regents of the
% University of California. The software program and documentation are 
% supplied "as is", without any accompanying services from The Regents. The 
% Regents does not warrant that the operation of the program will be
% uninterrupted 
% or error-free. The end-user understands that the program was developed for 
% research purposes and is advised not to rely exclusively on the program for any 
% reason.
% 
% IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO
% ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
% CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING
% OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
% EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
% THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
% CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
% THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
% THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO
% PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
% MODIFICATIONS.
%
% This software was made with government support under 
% Office of Naval Research Award Number: N000140110403. 
% The government has certain rights in this software.



    if nargin < 7 | round(LINE_ENDS_IGNORE)~=LINE_ENDS_IGNORE
        error('Not enough inputs to safe trace.')
    end

	xdata = [];
	ydata = [];
    x = x0;
    y = y0;
    Start = cputime;    % May wrap around and cause early aborting before 1 second limit.
    switch length(varargin),
    case 2
        TimeLimit = varargin{1};
        DistLimit = varargin{2};
    case 1
        TimeLimit = varargin{1};
        DistLimit = Inf;
    case 0
        TimeLimit = .5;
        DistLimit = Inf;
    end
        
    while 1
        if No_next_pt(x,y) 
            endCondition = 'gap';       break
        elseif Multiple_paths(x,y)
            endCondition = 'endOverlap';   break
        elseif abs(x-x0) > DistLimit
            endCondition = 'distance';  break
        elseif  (cputime-Start) > TimeLimit
            endCondition = 'timeout';   break
        end
        [y_index, top, botm] = get_connected_pixels(x,y,Img, g_width, MERGE);  
        
        % Check for the start of an overlap.
        if abs(x-x0) > MERGE & is_split(x-RATE,y_index, Img, MERGE), 
            endCondition = 'overlap';  break
        else, 
            y = center_y(x, top, botm, Img);            
        end
        if x ~= x0  % Do NOT include point passed in.
            xdata = [xdata; x];
            ydata = [ydata; y];
        end
        
        try
            [x, y, dif] = get_next_pt(x+RATE,y,y_index, Img);
        catch, 
            endCondition = 'gap';   break
        end   % Generally to catch index out of bounds.
    end%END WHILE
    
    % Remove last point, if it's often not reliable.
    if length(xdata) > LINE_ENDS_IGNORE & length(ydata) > LINE_ENDS_IGNORE
        switch endCondition
        case {'gap', 'overlap'}
            ydata((end-LINE_ENDS_IGNORE+1):end) = [];
            xdata((end-LINE_ENDS_IGNORE+1):end) = [];
        otherwise
        end
    end

% --------------------------------------------------------------------
% CENTER_Y - center the y value in the middle of the line
function y = center_y(x, top, botm, Img)
    y = top+(botm-top)/2;

    
% --------------------------------------------------------------------
% NO_NEXT_PT - tells us if get_next_pt could not find the next point
function bool = No_next_pt(x, y)
    if isempty(x) | isempty(y) | isnan(x) | isnan(y)
        bool = 1; % Must use if statement for ORs to short circuit to avoid [].
    else bool = 0; end
    
% --------------------------------------------------------------------
% MULTIPLE_PATHS - Tells us if we should record a forking point.
function bool = Multiple_paths(x,y)
    bool = length(y) > 1;
    %::DEBUG::
    if length(y) ~= length(x)
        error('mismatch x y')
    end
    
% --------------------------------------------------------------------
% IS_SPLIT - simpler alternative to is_overlap.
function retval = is_split(at_x,y_index, Img, MERGE)
    x_range = max(1,(at_x-floor((MERGE-1)/2))):min(size(Img,2),(at_x+ceil((MERGE-1)/2)));
    Slice = sum(Img(y_index,x_range),2);       % Merge lines together
    retval = any(diff(find(Slice))>1);
% --------------------------------------------------------------------
% DISK - create a disk structuring element
function nhood = disk(radius)
    [xx,yy] = meshgrid(-radius:radius);
    nhood = xx.^2 + yy.^2 <= radius^2;
    
    
    