%TRACE_IT Main tracing algorithm
%   This algorithm uses a vectorized version of finding the middle of the 
%   line.  The bulk of the code comprises of heuristic checks to try to stay on
%   the line when dealing combinations of gaps, overlaps, and image noise.
%
%   Parameters:
%       image_h - handle to image to be traced.
%       g_width - estimated average line width. (vertical width)
%       digit_button_h - handle to button so we can stop if its not pressed anymore.
%       MAX_PERIOD - maximum period in seconds to guess when we're tracing the wrong line
%       ZERO_P - Polynomial coefficients for the baseline.
%       dpi - the dpi of the image for scaling certain parameters.
%
% Required functions: 
%   safe_trace.m
%   get_connected_pixels.m
%   get_next_pt.m

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Author: Steve Chuang
%
% This software is Copyright ? 2002 The Regents of the University of California. 
% All Rights Reserved.
% 
% Permission to use, copy, modify, and distribute this software and its 
% documentation for educational, research and non-profit purposes, without fee, 
% and without a written agreement is hereby granted, provided that the above 
% copyright notice, this paragraph and the following three paragraphs appear in 
% all copies.
% 
% Permission to incorporate this software into commercial products may be
% obtained by contacting Technology Transfer Office, 9500 Gilman Drive, 
% 411 University Center, University of California, La Jolla, CA 92093-0093
% (858) 534-5815, invent@ucsd.edu
% 
% This software program and documentation are copyrighted by The Regents of the
% University of California. The software program and documentation are 
% supplied "as is", without any accompanying services from The Regents. The 
% Regents does not warrant that the operation of the program will be
% uninterrupted 
% or error-free. The end-user understands that the program was developed for 
% research purposes and is advised not to rely exclusively on the program for any 
% reason.
% 
% IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO
% ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
% CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING
% OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
% EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
% THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
% CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
% THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
% THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO
% PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
% MODIFICATIONS.
%
% This software was made with government support under 
% Office of Naval Research Award Number: N000140110403. 
% The government has certain rights in this software.


function [TraceX, TraceY] = trace_it(image_h, xPts,yPts, g_width ...
,digit_button_h, MAX_PERIOD, ZERO_P, dpi, varargin)

% Error check
if length(xPts) < 1 | length(yPts) < 1yy
    helpdlg('Need at least two control points.','Trace error','modal')
%     disp('Need at least two control points.')
    return
end

% Ensure xPts is all integers
for I=1:length(xPts)
    xPts{I} = round(xPts{I});
end

% PARAMETERS:
g_width;
MAX_PERIOD;
LINE_SPACING = 80;
MINUTE_GAP = 50 * dpi/400;  % Scale gap to DPI
DEFAULT_MERGE = 1;          % Only one value of merge to average center value now used
LINE_ENDS_IGNORE = 4;       % End of lines are rounded. When jumping a gap, ignore last few values.

Merge = DEFAULT_MERGE;

Img = init_image(image_h);
TraceY = init_array(Img);
[x, y, pt_index] = init_pts(xPts,yPts); % Pt_index = 0;
[prev_x,prev_y,top,botm,radiusTop,radiusBotm] = init_empty;     % Empty vars, just allocate mem.

[lastCross, crossDist] = reset_cross_info(x, y, ZERO_P);        % Crossing baseline info.
lastOverlap = reset_lastOverlap;
currentState = 'default';
SPLINE_FLAG = -Inf;
RATE = 1;                       % Value greater than 1 untested.
RATE = max(1,round(RATE));
Flush_timer = reset_timer;      % Reset Flush_timer

if length(varargin) > 0
    debug_file = [varargin{1} '.txt'];
    FID = fopen(debug_file,'a');
else FID = [];
end

% MAIN LOOP
try, while not_last_ctrl_pt(pt_index, length(xPts))
    % Record last point.
    [prev_x,prev_y] = update_records(x,y);

    if at_next_ctrl_pt(x, pt_index+1, xPts, RATE)
        % 1. determine control point
        % 2. Reverse trace from start of that control point
        % 3. Spline through control point
        % 4. Set up next point to continue tracing from end of control point
        Merge = DEFAULT_MERGE;
        [xpt,ypt, pt_index] = get_next_ctrl_pt(xPts, yPts, pt_index+1);
        if isequal(currentState,'default')
            [xdata, ydata] = safe_trace(xpt(1), ypt(1), Img, -RATE, g_width, Merge, LINE_ENDS_IGNORE, Inf, Inf);
            TraceY(xdata) = ydata;
            [x,y,TraceY] = process_ctrl_pt(xpt, ypt, TraceY, SPLINE_FLAG);
            [y_index, top, botm] = get_connected_pixels(x,y, Img, g_width, Merge);
            
            % Clear in case we plotted a point forward and jumped back to the control point
            TraceY = clear_data_after(x,TraceY);
        else
            % If state isn't normal, leave the control pt for splining later.
            [ignore,ignore,TraceY] = process_ctrl_pt(xpt, ypt, TraceY, SPLINE_FLAG);
        end
        
        if x ~= round(x)
            warning('Non-integer position found. Rounding value.')
            x = round(x);
        end
    else
        % Get next point
        [x, y, dif] = get_next_pt(prev_x+RATE,prev_y,y_index, Img);
        
        % SPECIAL CONDITIONS:
        if No_next_pt(x,y)
            if ~isvalid(prev_x) | ~isvalid(prev_y), error('Case exception: No previous point recorded!'), end

            % Try jumping
            % 1. Jump a small distance to overcome noise
            % 2. Try jumping a minute mark
            % 3. Go back to mean and reverse trace
            % 4. Set spline flag if distance jumped is small
            % 5. Reset our cross base line check if we jumped far.
            
            H = max([20, round(getPk2Pk(prev_x,TraceY,MAX_PERIOD))]);
            Avg = polyval2(ZERO_P,x);
            if isnan(Avg)
                Avg = prev_y;
            end
            
            % Step 1:
            [x,y] = Jump_to_nearest(prev_x+1, prev_y, Img, H, MINUTE_GAP);   % Small jumps ony
            % Step 2:
            if ~isvalid(y)
                predict_func = getPredictionFunc(TraceY, prev_x);
                [x,y] = Jump_minute_mark(prev_x+1,prev_y,Img, Avg,predict_func,H,MINUTE_GAP); 
            end
            % Step 3:
            if ~isvalid(y)
                [x, y] = Go_back_to_mean(x, ZERO_P, Img); % Go back to mean
                [xdata, ydata] = safe_trace(x,y,Img, -RATE,g_width,Merge,LINE_ENDS_IGNORE, Inf, Inf);
                TraceY(xdata) = ydata;
            end
            % Error Check:
            if ~isvalid(y)
                % Could not recover line. Skipping to next control point.
                x = Inf; continue   % Infinite to flag that we have gone past the ctrl point
            else
                if isequal(currentState,'default')
                    [xdata, ydata] = safe_trace(xpt(1),ypt(1),Img, -RATE,g_width,Merge,LINE_ENDS_IGNORE, Inf, Inf);
                    TraceY(xdata) = ydata;
                    TraceY = clear_data_after(x,TraceY); 
                end
                currentState = 'Just_jumped';
                rounded_edge = (prev_x-LINE_ENDS_IGNORE+1):(prev_x);  % Set to be indices
                filtered_edge = [min(rounded_edge)-4:min(rounded_edge)-1];
                TraceY(rounded_edge) = SPLINE_FLAG;
                rounded_edge = length(rounded_edge);    % Now set this value to be a counter

                % Do some smoothing at beginning of gap
                TraceY(filtered_edge) = myFilter(TraceY(filtered_edge));
                filtered_edge = 4;
            end
            
            % Step 4:
            if x-prev_x <= MINUTE_GAP
                TraceY(prev_x+1:x-1) = SPLINE_FLAG; % Flag for later splining.
            end
            % Step 5:
            if Jumped_over_period(x, prev_x, MAX_PERIOD)   
                % Resetting cross info on jump.
                [lastCross, crossDist] = reset_cross_info(x,y,ZERO_P);
            end
        elseif Multiple_paths(x,y)
            % Pick a path
            avg = polyval2(ZERO_P,x);
            [xpt,ypt] = get_next_ctrl_pt(xPts, yPts, pt_index+1);
            y = rank_paths(x,y, y_index, xpt,ypt, lastOverlap.X, TraceY, dif, avg, Img, g_width);
            y = y(1);
            x = x(1);
            
%             % Test for unexpected fork
%             % If we did not detect the start of an overlap, 
%             % assume it was caused by one of the lines having a 
%             % gap.  Just spline through the minute gap region.
%             if isequal(currentState, 'default')
%                 % This does not insure there are end points to do a
%                 % B-spline through.
%                 TraceY(max(1,x-MINUTE_GAP):max(1,x-1)) = SPLINE_FLAG;
%             end
        end%END SPECIAL CONDITIONS
        
        % Catch when we can't find any pixels:
        if isempty(x) | isempty(y)
            x = Inf;
            currentState = 'default';
            continue;
        end
        
        % CHECK IF WE'RE LOST:
        % Update cross mean status
        [lastCross,newCrossDist] = update_cross_info(x,y,crossDist,lastCross,ZERO_P);
        if Off_mean(x,y,Img,lastCross,newCrossDist,MAX_PERIOD,LINE_SPACING)
            % Recover via the mean
            % Reverse trace
            % Reset state
            if ~isempty(FID)
                fprintf(FID, 'Going back to mean at (%d,%d)\n', x,y);
            end
            [x, y] = Go_back_to_mean(x, ZERO_P, Img);
            [lastCross,newCrossDist] = update_cross_info(x,y,crossDist,lastCross,ZERO_P);
        else
            crossDist = newCrossDist;
        end%END CROSS MEAN CHECK
        
        Merge = DEFAULT_MERGE;  % Legacy code, only one merge value used now.

        % WHITE AREA: Gather new white area data about this pixel
        try
            [y_index, top, botm] = get_connected_pixels(x,y,Img, g_width, Merge);
        catch
            error(lasterr)
        end

        % ADJUST Y and SET CURRENT STATE
        switch currentState
        case 'overlap'
            % Adjust the y value
            y = adjust_y(prev_y,top,botm, radiusTop, radiusBotm);
            
            % Check of end-of-overlap
            if is_overlap(x+RATE, y_index, Img, 1, Merge, g_width)
                % Set state
                currentState = 'default';
                LIMIT = 50 * dpi/400;   % Scale w/ DPI
				TraceY = fix_overlap_region(x, y_index, lastOverlap, TraceY, SPLINE_FLAG, LIMIT);
                y = SPLINE_FLAG;
            elseif False_overlap(x, lastOverlap.X, MAX_PERIOD)
                % Reset state
                % Go back to where we though there was an overlap
                % Reset data to cotinue trace as default state
                currentState = 'default';   % Setting state back to default guarantees progress.
                [x,y,y_index,top,botm] = restoreLastOverlap(lastOverlap, Img, g_width);      % Backtrack to this point. 
                [y, radiusTop, radiusBotm] = center_y(x, top, botm, Img);   % Default behavior
            end
        case {'default', 'Just_jumped'}
            % Check for the start of an overlap.
            if is_overlap(x-RATE, y_index, Img, -1, Merge, g_width)
                % Set overlap state and record position
                % Adjust y instead of centering it
                currentState = 'overlap';
                if ~isempty(radiusTop) & ~isempty(radiusBotm)
                    lastOverlap = record_last_overlap(x,top,botm);
                    y = adjust_y(prev_y,top,botm, radiusTop, radiusBotm);
                else
                    [y, radiusTop, radiusBotm] = center_y(x, top, botm, Img);
                end

                % RECORD data point
                try
                    TraceY(x) = y;
                catch
                    error(lasterr)
                end
            else
                % Center the y value in the middle of the line
                [y, radiusTop, radiusBotm] = center_y(x, top, botm, Img);
                if isequal(currentState, 'Just_jumped')
                    if rounded_edge > 0
                        TraceY(x) = SPLINE_FLAG;        % Ignore value, spline through
                        rounded_edge = rounded_edge-1;  % Decrement count of how many points ignored
                    elseif rounded_edge == 0 & filtered_edge > 0
                        filtered_edge = filtered_edge-1;
                        TraceY(x) = y;  % Data point for filtering later
                    else
                        % Smooth out a bit of data at end of gap
                        TraceY(x-4:x-1) = myFilter(TraceY(x-3:x));
                        TraceY(x) = y;
                        currentState = 'default';
                    end
                else
                    currentState = 'default';
                    % RECORD data point
                    try
                        TraceY(x) = y;                    
                    catch
                        error(lasterr)
                    end
                end
            end
        otherwise
            error('Unknown state')
        end
    end % End finding the next point.

    % Check if user wants to stop
    if cputime-Flush_timer > 1      % Flush every 1 seconds.
        drawnow                     % Flush.
        if get(digit_button_h,'Value') ~= get(digit_button_h,'Max')
            break
        end
        Flush_timer = cputime;      % Reset flush timer.
    end

end, catch  % End main while-loop
    error(['Trace error in main loop -> ' lasterr])
end

try%POST LOOP
    % Clear spline flags at beginning
    idx = 1;
    while TraceY(idx) == SPLINE_FLAG
        TraceY(idx) = NaN;
        idx = idx+1;
    end
    % Clear spline flags at end
    idx = length(TraceY);
    while TraceY(idx) == SPLINE_FLAG
        TraceY(idx) = NaN;
        idx = idx-1;
    end

    
	% Fill in ends of trace
	if ~not_last_ctrl_pt(pt_index, length(xPts))
        [xdata, ydata] = safe_trace(xpt(1), ypt(1), Img, +RATE, g_width, Merge, LINE_ENDS_IGNORE, Inf, Inf);
        TraceY(xdata) = ydata;
	end
        
	% Do final tasks before finishing
	try
        if ~isempty(FID)
            fclose(FID);
        end
	catch
	end
	
	% Filter to avoid sharp ends when splining - Also deals w/ noise.
	TraceY = myFilter(TraceY);          % Moving average low-pass filter.
	TraceY = restore_ctrl_pts(TraceY, [xPts{:}], [yPts{:}]);
	
	% Spline through any missing data
	xx = find(TraceY == SPLINE_FLAG);
	X = find(isvalid(TraceY));
	Y = TraceY(X);
	xx = sort(xx);
	if ~isempty(xx)
        yy = spline(X,Y,xx);            % Spline through all
        TraceY(xx) = yy;
	end
	
	XOFFSET = 0;
	TraceX = find(isvalid(TraceY)) + XOFFSET;
	TraceY = TraceY(isvalid(TraceY));
	
catch%POST LOOP
    error(['Trace error after main loop -> ' lasterr])
end

% --------------------------------------------------------------------
% **************** END MAIN FUNCTION *********************************
% --------------------------------------------------------------------

% --------------------------------------------------------------------
function TraceY = restore_ctrl_pts(TraceY, x_ctrl, y_ctrl)
    TraceY(x_ctrl) = y_ctrl;
    
% --------------------------------------------------------------------
% MYFILTER - moving average filter, avoiding NaNs.
function ydata = myFilter(ydata)
    Idx = [];               % Preallocate memory
    B = [0.25 0.5 0.25];    % Filter kernal with a 3-point filter
    
    Invalid = find(~isvalid(ydata));
    Valid = find(isvalid(ydata));
    next_bad = min(Invalid);
    n = min(Valid);
    while n <= length(ydata)
        if isempty(next_bad)
            next_bad = length(ydata)+1;
        end
        Idx = [];   % Reset
        Idx = n:(next_bad-1);
        if ~isempty(Idx)
            temp = ydata(Idx);
            temp = filter2(B,[temp(1) temp temp(end)]);     % Pad the ends
            ydata(Idx) = temp(2:end-1);                     % Remove padding
        end
        n = min(Valid(Valid > next_bad));
        if isempty(n)
            break
        end
        next_bad = min(Invalid(Invalid > n));
    end

% --------------------------------------------------------------------
% **************** INITITLIZE - Initialize stuff ****************
% --------------------------------------------------------------------
% ------------------------------------------------------------------------
% INIT_IMG - grabs entire image of hte handle
function Img = init_image(image_h)
	if ~ishandle(image_h)
        error('Cannot trace: bad image handle passed to trace_it.')
	end
    Img = get(image_h,'CData');                          % Initialize image
    
% -----------------------
% INIT_ARRAY - matches array size to width of image.
function TraceY = init_array(Img)
    % x coordinate in Img corresponds to index in TraceY.  Convenient, don't you think?
	TraceY = zeros(1, size(Img,2)); % Row vector. 
	TraceY(:) = NaN;                % Fill with NaNs

% -----------------------
% INIT_EMPTY
function varargout = init_empty(varargin)
   for n=1:nargout
       varargout{n} = [];
   end
% --------------------------------
% INIT_PTS
function [x, y, pt_index] = init_pts(xPts, yPts)
    [xPts, yPts] = sortCtrlPts(xPts,yPts);
    x = min(xPts{1});
    y = min(yPts{1});
    pt_index = 0;

    % -----------------------------------
    % SORTCTRLPTS - sort by min x value.
	function [xPts, yPts] = sortCtrlPts(xPts,yPts)
		for P = 2:length(xPts)     % Insertion Sort
			temp = xPts{P};
            tempY= yPts{P};
            % Error check
            if ~isvalid(temp) | ~isvalid(tempY), error('Invalid control point.'),temp,tempY,end
			j = P;
			while j > 1 & min(xPts{j-1}) > min(temp)    % sort by min value of vector.
                xPts{j} = xPts{j-1};
                yPts{j} = yPts{j-1};
                j = j-1;
			end
			xPts{j} = temp;
            yPts{j} = tempY;
		end
% --------------------------------
% RESET_TIMER
function Flush_timer = reset_timer;      % Reset Flush_timer
    Flush_timer = cputime;
        
        
        
% --------------------------------------------------------------------
% **************** OVERLAP - overlap detection stuff ****************
% --------------------------------------------------------------------
% --------------------------------------------------------------------
% IS_SPLIT - simpler alternative to is_overlap.
function retval = is_split(at_x,y_index, Img, direction, Merge)
    x_range = max(1,(at_x-floor((Merge-1)/2))):min(size(Img,2),(at_x+ceil((Merge-1)/2)));
    Slice = any(Img(y_index,x_range),2);       % Merge lines together
    retval = any(diff(find(Slice))>1);
% --------------------------------------------------------------------
% IS_OVERLAP - function to check for beginning or end of overlap
% Does some more seismograph specific checks for overlaps.
%  Look for split - Look TWO pixels forward or back to avoid 1 pixel noise!
function retval = is_overlap(at_x,y_index, Img, direction, Merge, g_width)
try
    if max(y_index)-min(y_index) < g_width
        retval = 0; % Can't be overlap if it's not thick.
        return
    end
    
    EXTEND = 1;         % Distance to look forward or backward to confirm overlap.
    if direction < 0
        x_index = max(at_x-EXTEND,1):min(at_x,size(Img,2));  % Must not go past at_x.
    else x_index = max(at_x,1):min(at_x+EXTEND,size(Img,2));
    end 
    MainSlice = double(Img(y_index,at_x));  % Main slice.

    retval = 0;
    split = is_split_slice(MainSlice);      % Check if multiple distinct lines
    if split
        dif = diff([0; MainSlice; 0]);
        flipB = find(dif < 0)-1;                  % 1 to 0
        flipW = find(dif > 0);                  % 0 to 1;
        if length(flipB) ~= length(flipW), 
            error('Mismatched matrix dimensions'), end
        Mids = round(flipW+(flipB-flipW)./2);   % Mid point of split
        MultSlice = Img(y_index,x_index);       % Multiple columns together.

        Extended = is_extended(Mids, MultSlice) ;
        if Extended
            if split
                blockSizes = getBlockSize(MainSlice);
                MIN_W = 3;  % Min line width, noise otherwise
                Thick = length(find(blockSizes >= MIN_W))>1;   % More than 1 thick line
                if Thick 
                    if direction < 0
                        retval = is_split_slice(MultSlice(:,1));
                    else retval = is_split_slice(MultSlice(:,end));
                    end
                end
            end
        end
    end
catch
    error(['is_overlap error: ' lasterr])
end
	% ------------------------------
    function blockSizes = getBlockSize(Slice)
        Slice = [0; Slice; 0];              % Pad with zeros 
        flips = find(diff(Slice)~=0);       % Index of changes in color
        blockSizes = diff(flips);           % Difference between indices
        blockSizes = blockSizes(1:2:end);   % Grab white blockSizes (evens)
	% ------------------------------
	function split = is_split_slice(Slice)  % Check if multiple distinct lines
        split = any(diff(find(Slice))>1);   % See if we have a solid white block
    % ------------------------------
	function bool = is_extended(Mids, MultSlice) 
        bool = length(get_extended(Mids, MultSlice)) > 1;
    % ------------------------------
    function Extended = get_extended(Mids, MultSlice)
        Extended = Mids(all(MultSlice(Mids,:),2)); % Line extends.
%         figure(2), imagesc(MultSlice)
%         figure(1), imagesc(Img(y_index,at_x-9:at_x+10))

% --------------------------------------------------------------------
% FIX_OVERLAP_REGION
function TraceY = fix_overlap_region(x, y_index, lastOverlap, TraceY, SPLINE_FLAG, LIMIT)
if abs(x-lastOverlap.X) < LIMIT
    TraceY(lastOverlap.X:x) = SPLINE_FLAG;
end
% --------------------------------------------------------------------
% --- RESET_LASTOVERLAP ---
function lastOverlap = reset_lastOverlap
    lastOverlap.X = [];
    lastOverlap.Y = [];
% --------------------------------------------------------------------
% --- RECORD_LAST_OVERLAP ---
function lastOverlap = record_last_overlap(x,top,botm)
    lastOverlap.X = x;       % Allow us to backtrack to this point.
    lastOverlap.Y = top+1;   % Pick any point on line.
% --------------------------------------------------------------------
% --- RESTORELASTOVERLAP - Backtrack to this point.
function [x,y, y_index, top, botm] = restoreLastOverlap(lastOverlap, Img, g_width)
    x= lastOverlap.X;
    y = lastOverlap.Y;
    [y_index, top, botm] = get_connected_pixels(x,y,Img,g_width,1);
% --------------------------------------------------------------------
% FALSE_OVERLAP - if we can't find end of overlap after half a period,
%  it must be a false alarm. Take appropriate action.
function retval = False_overlap(x,lastOverlapX,MAX_PERIOD)
    retval = (x-lastOverlapX) >= MAX_PERIOD;
    
% --------------------------------------------------------------------
% ***** EXAMINE - examine new point *******
% --------------------------------------------------------------------
% --------------------------------------------------------------------
% --- RANK the mid points -----
function y = rank_paths(x,y, y_index, xpt, ypt, lastOverlapX, TraceY, dif, avg, Img, g_width)
    Merge = 1;
    Multiplier = zeros(size(y));
    tr_avg = zeros(size(y));
    for n=1:length(y)
        x0 = x(n);
        y0 = y(n);
        % Look ahead to see which path
        [xdata, ydata, endCondition] = safe_trace(x0,y0,Img,+1,g_width, Merge, 0, 1.0, 20);
        switch endCondition
        case 'overlap'
            Multiplier(n) = 1;
        case 'distance'
            Multiplier(n) = 1;
        case 'gap'
            Multiplier(n) = 1;
        case 'timeout'
            Multiplier(n) = 1;
        case 'endOverlap';
            Multiplier(n) = 2;
        end
        Multiplier(n) = Multiplier(n)./log(length(ydata)+2);
        tr_avg(n) = nanmean([y; ydata(:)]);
    end
    
    blockSizes = getBlockSize(double(Img(y_index,unique(x))));
    if length(blockSizes) == length(y)
        Multiplier = Multiplier ./ (20*log(blockSizes+1)+1);
    end

    D_to_ctrlpt = (xpt(1)-x).^2 + (ypt(1)-y).^2;
    D_to_mean = (tr_avg-avg).^2+30^2;         % Not a big factor until more than 30 pixels
    Rank = Multiplier.*min(D_to_ctrlpt, D_to_mean);     % Take smallest from each group.

    y = sortrows([Rank(:) y(:)]);           % Sort by rank
    y = y(:,2);                             % Return only y-values.
 
    
% --------------------------------------------------------------------
% MULTIPLE_PATHS
function bool = Multiple_paths(x,y)
    bool = length(y) > 1;
    % Check for error
    if length(y) ~= length(x)
        error('mismatch x y')
    end
% --------------------------------------------------------------------
% NO_NEXT_PT - tells us if get_next_pt could not find the next point
function bool = No_next_pt(x, y)
    if isempty(x) | isempty(y) | isnan(x) | isnan(y)
        bool = 1; % Must use if statement for ORs to short circuit to avoid [].
    else bool = 0; end
% --------------------------------------------------------------------
% ***** JUMP - jumping gaps and recovering the line *****************
% --------------------------------------------------------------------
% -----------------------------------------------
% GETPREDICTFUNC - simple slope prediction
% x - predict points after this time.
function predict_func = getPredictionFunc(TraceY, x)
    if isempty(x)
        predict_func = NaN;  
    elseif x > 1
        y1 = TraceY(x-1);
        y2 = TraceY(x);
        
        % Check if valid
        if ~isvalid(y2)
            predict_func = nanmean(TraceY(isvalid(TraceY)));
        elseif ~isvalid(y1)
            predict_func = y2;
        else
            predict_func = polyfit([x-1, x],[y1, y2],1);
        end
    else
        predict_func = TraceY(x);
    end
% --------------------------------------------------------------------
% JUMPED_OVER_PERIOD
function bool = Jumped_over_period(x, prev_x, MAX_PERIOD)
    bool = x-prev_x > MAX_PERIOD;
% -----------------------------------------------
% GETWINDOW - grabs a window for jumping gaps.
function [Window, y_offset] = getWindow(Img, x, y, FORWARD, HEIGHT)
    HEIGHT = round(HEIGHT);
    FORWARD = round(FORWARD);
    y = round(y);
    topRow = max(1,y-HEIGHT);                   % Make sure index in bounds
    botmRow = min(size(Img,1),y+HEIGHT);        % Make sure index in bounds
    leftCol = x;                                % Does not necessarily move forward
    rightCol = min(size(Img,2),x+FORWARD);       % Make sure index in bounds
    Window = Img(topRow:botmRow,leftCol:rightCol);
    y_offset = topRow-1;
% --------------------------------------------------------------------
% JUMP_TO_NEAREST
function [next_x,next_y] = Jump_to_nearest(x, y, Img, HEIGHT, DIST)
    HEIGHT = HEIGHT*1.5;            % 50% leeway
    next_x = x+1;   next_y = NaN;   % Default values if no pt found.
    
    [Window, y_offset] = getWindow(Img, x, y, DIST, HEIGHT);
    winY = y - y_offset;                          % y pixel in Window's referece

    if ~any(Window(:))
        return
    end
    
    % Create an array, D, with all the distances from the point of interest to 
    % each white pixel.
    [R C] = find(Window);           % Find every white pixel
    pt = [winY 0];                  % Point of interest
    D_from_pt = sum(abs([R C]-pt(ones(size(R)),:)).^2,2); % Squared distance
    D = D_from_pt;                  % Find the nearest pixel
    
    minD = min(D);
    if minD > 10^2
        return
    end
     
    i = min(find(D == minD));     % Pick left-top most pixel in case of a tie.
    
    next_x = x+C(i)-1;              % Restore to whole image coordinates.
    next_y = y+R(i)-winY;

% --------------------------------------------------------------------
% JUMP_MINUTE_MARK - Designed to ignore the minute mark that may
%  exist with minute marks.
function [next_x,next_y] = Jump_minute_mark(x, y, Img, Avg, predict_func, HEIGHT, MAXGAP)
    HEIGHT = HEIGHT*1.5;    % 50% leeway
    R = round(Avg);
    next_x = x+1;   next_y = NaN;   % Values returned if nothing found.
    
    [Window, y_offset] = getWindow(Img, x, R, MAXGAP, HEIGHT);
    winY = y - y_offset;                          % y pixel in Window's referece
    winAvg = Avg - y_offset;
    
    if ~any(Window)
        % Jump gap did not find any points.
        return
    end
    
    siz = size(Window);
    
    % Two options: Jump to avoid minute mark blobs, or grab any pixel
    switch 'Ignore minute marks for now'
    case 'avoid minute mark blobs'
        r = [1:siz(1)   ones(1,ceil(siz(2)/2)+1) siz(1).*ones(1, ceil(siz(2)/2)+1)];
        c = [siz(2).*ones(1,siz(1)) floor(siz(2)/2):siz(2)  floor(siz(2)/2):siz(2)];
        Objects = bwselect(Window,c,r,8);
    otherwise
        Objects = Window;
    end

    if ~any(Objects)
        Objects = Window;
    end
    
    % Create an array, D, with all the distances from the point of interest to 
    % each white pixel.
    [R C] = find(Objects);                      % Find every white pixel
    Prediction = peval(predict_func, C+x-1)-y_offset;
    
    pt = [winY 0];                              % Point of interest
    D_from_pt = sum(abs([R C]-pt(ones(size(R)),:)).^2,2); % inline call to repmat
    D_from_prediction = 600+0.1*abs(R - Prediction);    % Vertical distance from prediction.
    D_from_avg = abs(R-winAvg);
    
    ROLL_OFF = MAXGAP / 2;                      % "Cut-off frequency"

    D = 1.5*(1+C) + (1+sqrt(D_from_pt./2))  + 0.5*(1+D_from_prediction) + D_from_avg;
    
    % Find the nearest pixel 
    i = min(find(D == min(D)));     % Pick left-top most pixel in case of a tie.
    
    next_x = x+C(i)-1;
    next_y = y+R(i)-winY;
    
    % Check if values within the image
    if next_x > size(Img,2) | next_x < 1
        next_x = size(Img,2);
    end
    if next_y > size(Img,1) | next_y < 1
        next_y = size(Img,1);
    end
    
% ------------------------------------------
% PLOT_STUFF
function plot_stuff(Window, R,C,Data, titleStr, F)
    Window = double(Window);
    for n=1:length(R)
        Window(R(n),C(n)) = Data(n);
    end
    figure(F), hold off, imagesc(Window), title(titleStr)
    colorbar

% ------------------------------------------
% PEVAL - evaluates the value from getPredictionFunc
function Prediction = peval(predict_func, xdata)
    if isnumeric(predict_func)
        Prediction = polyval(predict_func,xdata);
    else Prediction = feval(predict_func,xdata);    % Get predictions.
    end
% --------------------------------------------------------------------
% GO_BACK_TO_MEAN - always produces a value. May return value at edge of image.
function [x, y] = Go_back_to_mean(x, ZERO_P, Img)
    BLACK = 0;
    % y is found on the base line constrained by the boundaries of Img.
    y = min(max(1,round(polyval2(ZERO_P,x))),size(Img,1));  % Keep in bounds
    while x < size(Img,2) & Img(y,x) == BLACK,
        x = x+1;
        y = min(max(1,round(polyval2(ZERO_P,x))),size(Img,1));  % Keep in bounds
    end
    
% --------------------------------------------------------------------
% ***** CROSS INFO- stuff to check if we crossed the base line *******
% --------------------------------------------------------------------
% -------------------------------------------------------------------- 
% RESET_CROSS_INFO - crossing mean info
function [lastCross, crossDist] = reset_cross_info(x, y, ZERO_P)
    lastCross = x;                          % Worst-case just crossed mean
    crossDist = getCrossDist(x,y,ZERO_P);
% -------------------------------------------------------------------- 
% GETCROSSDIST
function crossDist = getCrossDist(x,y,ZERO_P);
    crossDist = y - polyval2(ZERO_P,x);       % Toggle as we go above and below
% -------------------------------------------------------------------- 
% UPDATE_CROSS_INFO - crossing mean info
function [lastCross,newCrossDist] = update_cross_info(x,y,crossDist,lastCross,ZERO_P)
    newCrossDist = getCrossDist(x,y,ZERO_P);
    if (crossDist>0) ~= (newCrossDist>0)
        lastCross = x;
    end
% -------------------------------------------------------------------- 
% OFF_MEAN
function bool = Off_mean(x,y,Img,lastCross,crossDist,MAX_PERIOD, LINE_SPACING)
    if abs(crossDist) < LINE_SPACING/1.5
        bool = 0;       % We're not off track if we're not far enough away from mean.
    else
        bool = Period_limit_exceeded(x,lastCross,MAX_PERIOD) | y < 1 | y > size(Img,1);
        if bool
            %disp('Off mean ')
        end
    end
% -------------------------------------------------------------------- 
% PERIOD_LIMIT_EXCEEDED
function bool = Period_limit_exceeded(x,lastCross,MAX_PERIOD)
    bool = x-lastCross >= MAX_PERIOD;
% -------------------------------------------------------------------- 
% POLYVAL2 - 1 or 2 coefficients only.
function y = polyval2(ZERO_P, x);
    switch length(ZERO_P) 
    case 1, y = ZERO_P;
    case 2, y = ZERO_P(1)*x + ZERO_P(2);
    otherwise
        ZERO_P
        error('POLYVAL2 error: Invalid line polynomial.')
    end
    
% --------------------------------------------------------------------
% ***** CONTROL POINT - get & process control points **************
% --------------------------------------------------------------------
% --------------------------------------------------------------------
% NOT_LAST_CTRL_PT
function bool = not_last_ctrl_pt(pt_index, numCtrlPts)
    bool = pt_index < numCtrlPts;
% --------------------------------------------------------------------
% GET_NEXT_CTRL_PT 
function [xpt,ypt, pt_index, TraceY] = get_next_ctrl_pt(xPts, yPts, next_pt_index);
    xpt = xPts{next_pt_index};
    ypt = yPts{next_pt_index};
    pt_index = next_pt_index;
% --------------------------------------------------------------------
% PROCESS_CTRL_PT - Fill in some values in trace and setup next ctrl point.
function [x,y,TraceY] = process_ctrl_pt(xpt, ypt, TraceY, SPLINE_FLAG);
    idx = find(xpt == max(xpt));    % Next point is at end of control point.
    x = xpt(idx);
    y = ypt(idx);

    xx = (min(xpt)+1):(max(xpt)-1); % Flag region for splining.
    TraceY(xx) = SPLINE_FLAG;       % Flag region for splining.
    TraceY(xpt) = ypt;              % Insert values we know.
% --------------------------------------------------------------------
% AT_NEXT_CTRL_PT
function bool = at_next_ctrl_pt(x, next_pt_index, xPts, RATE)
    if next_pt_index > length(xPts)
        error('No next control point.')
    end
    bool = (x+RATE) >= min(xPts{next_pt_index});


% --------------------------------------------------------------------
% ***** MISC - miscellaneous ***********************************
% --------------------------------------------------------------------
% --------------------------------------------------------------------
% CENTER_Y - center the y value in the middle of the line
function [y, radiusTop, radiusBotm] = center_y(x, top, botm, Img)
    y = top+(botm-top)/2;   % 0.6 seconds for 10,000 iterations. 
    radiusTop = y-top;      % Approx. thickness of top half of line.
    radiusBotm = botm-y;    % Approx. thickness of bottom half of line.
% --------------------------------------------------------------------
% ADJUST_Y - adjust the y value on an overlapped line
function y = adjust_y(prev_y, top, botm, radiusTop, radiusBotm)
    T = min(botm-1,top+radiusTop);    % Start from top, measure towards bottom. Can't go below botm.
    B = max(top+1,botm-radiusBotm);   % Start from bottom, measure towards top. Can't go above top.
    if abs(prev_y-T) < abs(prev_y-B)
        y = T;
    else y = B;
    end
    
% --------------------------------------------------------------------
% --- ISVALID ---
function boolean = isvalid(y)
    % Does empty have to be a separate test? I don't think so anymore, but I thought so before.
    boolean = ~isempty(y) & ~isnan(y) & isfinite(y) & y > 0;

% --------------------------------------------------------------------
% --- CLEAR_DATA_AFTER ---
function TraceY = clear_data_after(x, TraceY)
    TraceY(x+1:end) = NaN;
% --------------------------------------------------------------------
% --- UPDATE_RECORDS ---
function [prev_x,prev_y] = update_records(x,y)
    if ~isvalid(y) | ~isvalid(x)
        %warning('Invalid values traced. Trace runs off image or overlap spline?')
    end
    prev_x = x;         prev_y = y;         % Previous loop
% --------------------------------------------------------------------
% GETPK2PK - Returns the recent max amplitude
% Future revisions might apply a low pass filter and then find
% inflection points.  Return empty if not good samples found.
function pkpk = getPk2Pk(x, TraceData, DIST)
    Sample = TraceData(max(1,x-fix(DIST)):end);
    Sample = Sample(isvalid(Sample));   % Get only valid Sample points
    
    if isempty(Sample)  % Allow empty to pass through
        % No pts found to estimate pk2pk
    elseif any(~isvalid(Sample))
        error('Invalid samples getPk2Pk')
    end
	pkpk = max(Sample)-min(Sample);

